import fs from 'node:fs';
import path from 'node:path';
import { unified } from 'unified';
import { doi } from 'doi-utils';
import { VFile } from 'vfile';
import yaml from 'js-yaml';
import { copyNode, fileError, RuleId, normalizeLabel } from 'myst-common';
import { select, selectAll } from 'unist-util-select';
import { u } from 'unist-builder';
import { RefType } from 'jats-tags';
import { Jats } from 'jats-xml';
import { MathMLToLaTeX } from 'mathml-to-latex';
import { js2xml } from 'xml-js';
import { basicTransformations, citationToMixedCitation, journalTransforms, } from './transforms/index.js';
import { abstractTransform, descriptionFromAbstract } from './transforms/abstract.js';
import { processJatsReferences, resolveJatsCitations } from './transforms/references.js';
import { backToBodyTransform } from './transforms/footnotes.js';
import version from './version.js';
import { logMessagesFromVFile, toText } from './utils.js';
import { inlineCitationsTransform } from './myst/inlineCitations.js';
import { abbreviationSectionTransform } from './transforms/abbreviations.js';
import { floatToEndTransform } from './transforms/supplementary.js';
function refTypeToReferenceKind(kind) {
    switch (kind) {
        case RefType.sec:
            return 'heading';
        case RefType.fig:
            return 'figure';
        case RefType.dispFormula:
            return 'equation';
        case RefType.table:
            return 'table';
        case RefType.custom:
            return undefined;
    }
}
function texMathFromNode(node) {
    var _a, _b;
    const texMath = select('tex-math', node);
    if (texMath && ((_a = texMath.children) === null || _a === void 0 ? void 0 : _a[0].cdata)) {
        return (_b = texMath.children) === null || _b === void 0 ? void 0 : _b[0].cdata;
    }
    selectAll('*', node).forEach((n) => {
        if (n.type.startsWith('mml:')) {
            n.type = n.type.substring(4);
        }
    });
    const math = select('math', node);
    if (!math)
        return;
    [math, ...selectAll('math *', node)].forEach((n) => {
        const { type, value, children, ...attributes } = n;
        if (type === 'text') {
            n.type = 'text';
            n.text = value;
            delete n.value;
        }
        else {
            n.type = 'element';
            n.name = type;
            n.elements = children;
            n.attributes = attributes;
            delete n.children;
            Object.keys(attributes).forEach((k) => {
                delete n[k];
            });
        }
    });
    return MathMLToLaTeX.convert(js2xml({ type: 'element', name: 'root', elements: [math] }));
}
const handlers = {
    body(node, state) {
        state.renderChildren(node);
    },
    text(node, state) {
        state.text(node.value);
    },
    p(node, state) {
        state.renderInline(node, 'paragraph');
    },
    heading(node, state) {
        state.renderInline(node, 'heading', {
            enumerated: true,
            label: node.id,
            identifier: node.id,
            depth: node.depth,
        });
    },
    block(node, state) {
        var _a;
        state.renderInline(node, 'block', { data: { part: (_a = node.part) !== null && _a !== void 0 ? _a : node['sec-type'] } });
    },
    ['disp-quote'](node, state) {
        state.renderInline(node, 'blockquote', { kind: node['content-type'] });
    },
    // definitionList(node, state) {
    //   state.renderInline(node, 'def-list');
    // },
    // definitionItem(node, state) {
    //   state.renderInline(node, 'def-item');
    // },
    // definitionTerm(node, state) {
    //   state.renderInline(node, 'term');
    // },
    // definitionDescription(node, state) {
    //   state.renderInline(node, 'def');
    // },
    // code(node, state) {
    //   const { lang } = node as Code;
    //   state.renderInline(node, 'code', { language: lang });
    // },
    list(node, state) {
        // https://jats.nlm.nih.gov/archiving/tag-library/1.3/element/list.html
        state.renderInline(node, 'list', {
            ordered: node['list-type'] === 'ordered',
        });
    },
    ['list-item'](node, state) {
        state.renderInline(node, 'listItem');
    },
    ['inline-formula'](node, state) {
        const texMath = texMathFromNode(node);
        if (texMath) {
            state.addLeaf('inlineMath', {
                value: texMath,
                label: node.id,
                identifier: node.id,
            });
        }
        else {
            state.renderChildren(node);
        }
    },
    ['disp-formula'](node, state) {
        const texMath = texMathFromNode(node);
        if (texMath) {
            state.addLeaf('math', {
                value: texMath,
                label: node.id,
                identifier: node.id,
            });
        }
        else {
            state.renderChildren(node);
        }
    },
    bold(node, state) {
        state.renderInline(node, 'strong');
    },
    italic(node, state) {
        state.renderInline(node, 'emphasis');
    },
    underline(node, state) {
        state.renderInline(node, 'underline');
    },
    monospace(node, state) {
        state.renderInline(node, 'inlineCode');
    },
    sub(node, state) {
        state.renderInline(node, 'subscript');
    },
    sup(node, state) {
        state.renderInline(node, 'superscript');
    },
    strike(node, state) {
        state.renderInline(node, 'delete');
    },
    sc(node, state) {
        state.renderInline(node, 'smallcaps');
    },
    // break(node, state) {
    //   // https://jats.nlm.nih.gov/archiving/tag-library/1.3/element/break.html
    //   state.addLeaf('break');
    // },
    // // abbreviation(node, state) {
    // //   // TODO: \newacronym{gcd}{GCD}{Greatest Common Divisor}
    // //   // https://www.overleaf.com/learn/latex/glossaries
    // //   state.renderChildren(node, true);
    // // },
    ['ext-link'](node, state) {
        state.renderInline(node, 'link', { url: node['xlink:href'] });
    },
    uri(node, state) {
        state.renderInline(node, 'link', { url: node['xlink:href'] });
    },
    ['boxed-text'](node, state) {
        state.renderInline(node, 'admonition', { kind: 'info' });
    },
    admonitionTitle(node, state) {
        // This is created in a transform!
        state.renderInline(node, 'admonitionTitle');
    },
    // attrib(node, state) {
    //   // This is used inside of disp-quotes
    //   state.renderInline(node, 'attrib');
    // },
    // image(node, state) {
    //   if (node.url?.startsWith('http')) {
    //     state.warn(`Image URL is remote (${node.url})`, node, 'image');
    //   }
    //   if (state.data.isInContainer && node.alt) {
    //     state.openNode('alt-text');
    //     state.text(node.alt);
    //     state.closeNode();
    //   }
    //   // TOOD: identifier?
    //   state.addLeaf('graphic', { 'xlink:href': node.url });
    // },
    ['fig-group'](node, state) {
        var _a;
        state.openNode('tabSet');
        (_a = node.children) === null || _a === void 0 ? void 0 : _a.forEach((n) => {
            state.openNode('tabItem', {
                title: toText(select('label', n)),
                sync: toText(select('label', n)),
            });
            state.renderChildren({ children: [n] });
            state.closeNode();
        });
        state.closeNode();
    },
    graphic(node, state) {
        const link = node === null || node === void 0 ? void 0 : node['xlink:href'];
        state.addLeaf('image', { url: link });
    },
    fig(node, state) {
        var _a;
        const caption = select('caption', node);
        const graphic = select('graphic', node);
        const title = select('title', node);
        const { label, identifier } = (_a = normalizeLabel(node.id)) !== null && _a !== void 0 ? _a : {};
        state.openNode('container', { label, identifier, kind: 'figure' });
        const wasInContainer = state.data.isInContainer;
        state.data.isInContainer = true;
        const link = graphic === null || graphic === void 0 ? void 0 : graphic['xlink:href'];
        if (link) {
            state.addLeaf('image', { url: link });
        }
        state.openNode('caption');
        if (title) {
            state.openNode('strong');
            state.renderChildren(title);
            state.closeNode();
        }
        // caption number?
        if (caption) {
            state.renderChildren(caption);
        }
        state.closeNode();
        state.closeNode();
        state.data.isInContainer = wasInContainer;
    },
    ['table-wrap'](node, state) {
        var _a;
        const captionNode = select('caption', node);
        const labelNode = select('label', node);
        const titleNode = select('title', node);
        const { label, identifier } = (_a = normalizeLabel(node.id)) !== null && _a !== void 0 ? _a : {};
        state.openNode('container', { label, identifier, kind: 'table' });
        const wasInContainer = state.data.isInContainer;
        state.data.isInContainer = true;
        state.openNode('caption');
        if (titleNode) {
            state.openNode('strong');
            state.renderChildren(titleNode);
            state.closeNode();
            titleNode.type = '__ignore__';
        }
        if (captionNode || labelNode) {
            state.renderChildren(captionNode || labelNode);
            if (captionNode)
                captionNode.type = '__ignore__';
            if (labelNode)
                labelNode.type = '__ignore__';
        }
        state.closeNode();
        state.renderChildren(node);
        state.closeNode();
        state.data.isInContainer = wasInContainer;
    },
    table(node, state) {
        state.openNode('table');
        state.renderChildren(node);
        state.closeNode();
    },
    thead(node, state) {
        state.renderChildren(node);
    },
    tbody(node, state) {
        state.renderChildren(node);
    },
    tfoot(node, state) {
        state.renderChildren(node);
    },
    tr(node, state) {
        state.openNode('tableRow');
        state.renderChildren(node);
        state.closeNode();
    },
    th(node, state) {
        const { align, colspan, rowspan } = node;
        state.openNode('tableCell', { header: true, align, colspan, rowspan });
        state.renderChildren(node);
        state.closeNode();
    },
    td(node, state) {
        const { align, colspan, rowspan } = node;
        state.openNode('tableCell', { align, colspan, rowspan });
        state.renderChildren(node);
        state.closeNode();
    },
    ['table-wrap-foot'](node, state) {
        state.openNode('legend');
        state.renderChildren(node);
        state.closeNode();
    },
    hr(node, state) {
        if (state.data.isInContainer)
            return;
        state.addLeaf('thematicBreak');
    },
    alternatives(node, state) {
        var _a;
        const choice = (_a = node.children) === null || _a === void 0 ? void 0 : _a.find((child) => !!state.handlers[child.type]);
        if (!choice) {
            state.error(`No supported types in 'alternatives' node`);
        }
        else {
            state.handlers[choice.type](choice, state, node);
        }
    },
    break(node, state) {
        state.addLeaf('break');
    },
    ['named-content'](node, state) {
        // TODO: Not just ignore things marked as named-content
        state.renderChildren(node);
    },
    ['fn-group'](node, state) {
        state.renderChildren(node);
    },
    fn(node, state) {
        var _a;
        const { label, identifier } = (_a = normalizeLabel(node.id)) !== null && _a !== void 0 ? _a : {};
        state.openNode('footnoteDefinition', { label, identifier });
        state.renderChildren(node);
        state.closeNode();
    },
    xref(node, state) {
        var _a;
        const refType = node['ref-type'];
        const { label, identifier } = (_a = normalizeLabel(node.rid)) !== null && _a !== void 0 ? _a : {};
        switch (refType) {
            case RefType.bibr:
            case RefType.ref:
                state.renderInline(node, 'cite', {
                    label,
                    identifier,
                    kind: 'narrative',
                });
                return;
            case RefType.supplementaryMaterial:
            case RefType.sec:
            case RefType.fig:
            case RefType.dispFormula:
            case RefType.table: {
                const kind = refTypeToReferenceKind(refType);
                state.renderInline(node, 'crossReference', { label, identifier, kind });
                return;
            }
            case RefType.fn:
            case RefType.tableFn: {
                state.addLeaf('footnoteReference', { label, identifier });
                return;
            }
            default: {
                state.renderInline(node, 'crossReference', { identifier: node.rid });
                state.warn(`Unknown ref-type of ${refType}`);
                return;
            }
        }
    },
    ['supplementary-material'](node, state) {
        var _a;
        if (node.id) {
            const { label, identifier } = (_a = normalizeLabel(node.id)) !== null && _a !== void 0 ? _a : {};
            state.openNode('div', { label, identifier });
        }
        state.renderChildren(node);
        if (node.id) {
            state.closeNode();
        }
    },
    media(node, state) {
        state.renderInline(node, 'link', { url: node['xlink:href'] });
    },
    ['inline-supplementary-material'](node, state) {
        state.renderInline(node, 'link', { url: node['xlink:href'] });
    },
    caption(node, state) {
        state.renderChildren(node);
    },
    // These nodes can be safely ignored
    label() { },
    comment() { },
    __ignore__() { },
    ['object-id'](node, state) {
        // TODO: handle other id types?
        if (node['pub-id-type'] === 'doi') {
            state.top().doi = toText(node);
        }
    },
};
const DEFAULT_HANDLERS = { ...handlers };
export class JatsParser {
    constructor(file, jats, opts) {
        var _a;
        this.stack = [];
        this.unhandled = [];
        this.file = file;
        this.jats = jats;
        this.options = opts !== null && opts !== void 0 ? opts : {};
        this.data = {};
        this.stack = [{ type: 'root', children: [] }];
        this.handlers = (_a = opts === null || opts === void 0 ? void 0 : opts.handlers) !== null && _a !== void 0 ? _a : DEFAULT_HANDLERS;
    }
    top() {
        return this.stack[this.stack.length - 1];
    }
    warn(message, source, opts) {
        fileError(this.file, message, {
            ...opts,
            source: source ? `jats-convert:${source}` : 'jats-convert',
            ruleId: RuleId.jatsParses,
        });
    }
    error(message, source, opts) {
        fileError(this.file, message, {
            ...opts,
            source: source ? `jats-convert:${source}` : 'jats-convert',
            ruleId: RuleId.jatsParses,
        });
    }
    pushNode(el) {
        var _a;
        const top = this.top();
        if (this.stack.length && el && 'children' in top)
            (_a = top.children) === null || _a === void 0 ? void 0 : _a.push(el);
        return el;
    }
    text(text) {
        var _a, _b;
        const top = this.top();
        const value = text;
        if (!value || !this.stack.length || !('children' in top))
            return;
        const last = (_a = top.children) === null || _a === void 0 ? void 0 : _a[top.children.length - 1];
        if ((last === null || last === void 0 ? void 0 : last.type) === 'text') {
            // The last node is also text, merge it
            last.value += `${value}`;
            return last;
        }
        const node = u('text', value);
        (_b = top.children) === null || _b === void 0 ? void 0 : _b.push(node);
        return node;
    }
    renderChildren(node) {
        var _a;
        (_a = node.children) === null || _a === void 0 ? void 0 : _a.forEach((child) => {
            const handler = this.handlers[child.type];
            if (handler) {
                handler(child, this, node);
            }
            else {
                this.unhandled.push(child.type);
                this.error(`Unhandled JATS conversion for node of "${child.type}"`);
            }
        });
    }
    renderInline(node, name, attributes) {
        this.openNode(name, { ...attributes });
        if ('children' in node) {
            this.renderChildren(node);
        }
        else if ('value' in node && node.value) {
            this.text(node.value);
        }
        this.closeNode();
    }
    addLeaf(name, attributes) {
        this.openNode(name, attributes, true);
        this.closeNode();
    }
    openNode(name, attributes, isLeaf = false) {
        const node = { type: name, ...attributes };
        if (!isLeaf)
            node.children = [];
        this.stack.push(node);
    }
    closeNode() {
        const node = this.stack.pop();
        return this.pushNode(node);
    }
}
export const jatsConvertPlugin = function (jats, opts) {
    this.Compiler = (node, file) => {
        if (jats.abstract)
            abstractTransform(jats.abstract);
        const tree = jats.abstract
            ? {
                type: 'root',
                children: [
                    u('block', { part: 'abstract' }, copyNode(jats.abstract).children),
                    ...copyNode(node).children,
                ],
            }
            : copyNode(node);
        // Can do better than this in the future, but for now, just put them at the end!
        const floatsGroup = selectAll('floats-group', jats.tree);
        if (floatsGroup.length > 0) {
            floatsGroup.forEach((g) => {
                tree.children.push(...g.children);
            });
        }
        basicTransformations(tree, file);
        citationToMixedCitation(tree);
        journalTransforms(tree);
        const state = new JatsParser(file, jats, opts !== null && opts !== void 0 ? opts : { handlers });
        state.renderChildren(tree);
        while (state.stack.length > 1)
            state.closeNode();
        if (state.unhandled.length && (opts === null || opts === void 0 ? void 0 : opts.logInfo)) {
            opts.logInfo.unhandled = [...new Set(state.unhandled)];
        }
        const referenceData = Object.fromEntries(jats.references.map((bibr) => {
            const id = bibr.id;
            const names = selectAll('name,string-name', bibr)
                .map((n) => `${toText(select('surname', n))}, ${toText(select('given-names', n))}`)
                .join(', ');
            const year = toText(select('year', bibr));
            const title = toText(select('article-title', bibr));
            const source = toText(select('source', bibr));
            const volume = toText(select('volume', bibr));
            const fpage = toText(select('fpage', bibr));
            const lpage = toText(select('lpage', bibr));
            const doiElement = selectAll('ext-link,[pub-id-type=doi]', bibr).find((e) => doi.validate(toText(e)));
            const doiString = doiElement ? toText(doiElement) : undefined;
            const doiLink = doiString ? ` <a href=${doi.buildUrl(doiString)}>${doiString}</a>` : '';
            return [
                id,
                {
                    html: `${names}. (${year}). ${title}. <i>${source}</i>, <i>${volume}</i>, ${fpage}-${lpage}.${doiLink}`,
                    doi: doiString,
                },
            ];
        }));
        const referenceOrder = [];
        const xrefs = selectAll('xref[ref-type=bibr]', jats.body);
        xrefs.forEach((xref) => {
            const rid = xref.rid;
            if (!referenceOrder.includes(rid)) {
                referenceOrder.push(rid);
            }
        });
        const result = {
            references: { order: referenceOrder, data: referenceData },
            tree: state.stack[0],
        };
        file.result = result;
        return file;
    };
    return (node) => {
        return node;
    };
};
export async function jatsConvertTransform(data, opts) {
    var _a, _b, _c;
    const jats = typeof data === 'string' ? new Jats(data) : data;
    if (opts === null || opts === void 0 ? void 0 : opts.logInfo) {
        opts.logInfo.publisher = toText(select('publisher-name', jats.tree)) || null;
        opts.logInfo.journal = toText(select('journal-title', jats.tree)) || null;
        opts.logInfo.pmid = toText(select('article-id[pub-id-type=pmid]', jats.tree)) || null;
        opts.logInfo.pmc = toText(select('article-id[pub-id-type=pmc]', jats.tree)) || null;
        opts.logInfo.doi = toText(select('article-id[pub-id-type=doi]', jats.tree)) || null;
        opts.logInfo.year = toText(select('year', jats.publicationDate)) || null;
        const license = select('license', jats.tree);
        let licenseString = null;
        if (license === null || license === void 0 ? void 0 : license['xlink:href']) {
            licenseString = license['xlink:href'];
        }
        else if (select('[type=ali\\:license_ref]', license)) {
            licenseString = toText(select('[type=ali\\:license_ref]', license));
        }
        else if (selectAll('ext-link', license).length === 1) {
            licenseString = (_a = select('ext-link', license)['xlink:href']) !== null && _a !== void 0 ? _a : null;
        }
        else if (license) {
            licenseString = toText(license);
        }
        opts.logInfo.license = licenseString;
    }
    const { frontmatter } = jats;
    const file = (_b = opts === null || opts === void 0 ? void 0 : opts.vfile) !== null && _b !== void 0 ? _b : new VFile();
    const refLookup = await processJatsReferences(jats, opts);
    floatToEndTransform(jats);
    backToBodyTransform(jats);
    const pipe = unified().use(jatsConvertPlugin, jats, opts);
    pipe.stringify(((_c = jats.body) !== null && _c !== void 0 ? _c : { type: 'body', children: [] }), file);
    const references = file.result.references;
    const tree = file.result.tree;
    resolveJatsCitations(tree, refLookup);
    inlineCitationsTransform(tree, [...Object.keys(references.data)]);
    abbreviationSectionTransform(tree, frontmatter);
    const abstract = selectAll('block', tree).find((node) => {
        return node.data && node.data.part === 'abstract';
    });
    if (abstract) {
        frontmatter.description = descriptionFromAbstract(toText(abstract));
    }
    if (opts === null || opts === void 0 ? void 0 : opts.logInfo) {
        opts.logInfo.figures = {
            body: selectAll('fig', jats.body).length,
            back: selectAll('fig', jats.back).length,
            myst: selectAll('container[kind=figure]', tree).length,
        };
        opts.logInfo.tables = {
            body: selectAll('table-wrap', jats.body).length,
            back: selectAll('table-wrap', jats.back).length,
            myst: selectAll('container[kind=table]', tree).length,
        };
        opts.logInfo.math = {
            inline: {
                body: selectAll('inline-formula', jats.body).length,
                back: selectAll('inline-formula', jats.back).length,
                myst: selectAll('inlineMath', tree).length,
            },
            equations: {
                body: selectAll('disp-formula', jats.body).length,
                back: selectAll('disp-formula', jats.back).length,
                myst: selectAll('math', tree).length,
            },
        };
        opts.logInfo.footnotes = {
            body: selectAll('fn', jats.body).length,
            back: selectAll('fn', jats.back).length,
            myst: selectAll('footnoteDefinition', tree).length,
        };
    }
    return { tree, jats, references, frontmatter }; //, kind };
}
export async function jatsConvert(session, input, opts) {
    const logInfo = { jatsVersion: version };
    const dir = path.dirname(input);
    const vfile = new VFile();
    vfile.path = input;
    const { tree, frontmatter } = await jatsConvertTransform(fs.readFileSync(input).toString(), {
        vfile,
        dir,
        logInfo,
        dois: opts === null || opts === void 0 ? void 0 : opts.dois,
        bibtex: opts === null || opts === void 0 ? void 0 : opts.bibtex,
    });
    logMessagesFromVFile(session, vfile);
    const basename = path.basename(input, path.extname(input));
    const mystJson = path.join(dir, `${basename}.myst.json`);
    const mystYml = path.join(dir, 'myst.yml');
    const logJson = path.join(dir, `${basename}.log.json`);
    const logYml = path.join(dir, `${basename}.log.yml`);
    fs.writeFileSync(logJson, JSON.stringify(logInfo, null, 2));
    fs.writeFileSync(logYml, yaml.dump(logInfo));
    if ((opts === null || opts === void 0 ? void 0 : opts.frontmatter) === 'page') {
        fs.writeFileSync(mystJson, JSON.stringify({ mdast: tree, frontmatter }, null, 2));
    }
    else if ((opts === null || opts === void 0 ? void 0 : opts.frontmatter) === 'project') {
        if (fs.existsSync(mystYml)) {
            // console.log('myst.yml exists; overriding with frontmatter from JATS');
            const previous = yaml.load(fs.readFileSync(mystYml).toString());
            fs.writeFileSync(mystYml, yaml.dump({ ...previous, project: { ...previous.project, ...frontmatter } }));
        }
        else {
            // console.log(`writing new myst.yml file`);
            fs.writeFileSync(mystYml, yaml.dump({ version: 1, project: frontmatter, site: {} }));
        }
        fs.writeFileSync(mystJson, JSON.stringify({ mdast: tree, frontmatter: { title: frontmatter.title } }, null, 2));
    }
    else {
        // console.log(`ignoring frontmatter`);
        fs.writeFileSync(mystJson, JSON.stringify({ mdast: tree, frontmatter: { title: frontmatter.title } }, null, 2));
    }
}
