import { copyNode, liftChildren, normalizeLabel } from 'myst-common';
import { remove } from 'unist-util-remove';
import { selectAll } from 'unist-util-select';
/**
 * Remove cite node children
 *
 * These children should be auto-computed by MyST
 */
function removeCiteChildren(tree) {
    const cites = selectAll('cite', tree);
    cites.forEach((cite) => {
        delete cite.children;
    });
}
/**
 * Ensure all cite nodes are nested in a citeGroup
 *
 * Existing citeGroups are unchanged; standalone cite nodes
 * become citeGroups with a single child.
 *
 * This function is run to simplify further inline citation transforms
 * so we only need to worry about the citeGroup case, not the
 * standalone cite case.
 */
function allCitesToCiteGroups(tree) {
    const standaloneCites = selectAll(':not(citeGroup) > cite', tree);
    standaloneCites.forEach((node) => {
        const cite = copyNode(node);
        delete node.enumerator;
        delete node.error;
        delete node.identifier;
        delete node.label;
        delete node.partial;
        delete node.prefix;
        delete node.suffix;
        const citeGroup = node;
        citeGroup.type = 'citeGroup';
        citeGroup.children = [cite];
    });
}
/**
 * Ensure all citeGroups do not contain citeGroups
 *
 * This unlikely scenario may only be achieved when a JATS citation
 * that is already part of a citeGroup points to multiple references
 * and is resolved to become a citeGroup itself.
 */
function flattenNestedCiteGroups(tree) {
    const nestedCiteGroups = selectAll('citeGroup > citeGroup', tree);
    nestedCiteGroups.forEach((node) => {
        node.type = '__lift__';
    });
    liftChildren(tree, '__lift__');
}
/**
 * Combine adjacent citeGroups into single citeGroup
 *
 * This function applies even if the citeGroups have a different kind
 * the kind of the first will be maintained.
 */
function combineAdjacentCiteGroups(tree) {
    const citeGroupParents = selectAll(':has(> citeGroup)', tree);
    citeGroupParents.forEach((parent) => {
        parent.children.forEach((child, index) => {
            var _a, _b;
            if (child.type !== 'citeGroup')
                return;
            const nextChild = parent.children[index + 1];
            if ((nextChild === null || nextChild === void 0 ? void 0 : nextChild.type) !== 'citeGroup')
                return;
            nextChild.children = [...((_a = child.children) !== null && _a !== void 0 ? _a : []), ...((_b = nextChild.children) !== null && _b !== void 0 ? _b : [])];
            nextChild.kind = child.kind;
            child.type = '__delete__';
        });
    });
    remove(tree, '__delete__');
}
/**
 * Remove commas and semicolons between citeGroup nodes and combine to single citeGroup
 */
function removeCiteSeparators(tree) {
    const citeGroupParents = selectAll(':has(> citeGroup + text + citeGroup)', tree);
    citeGroupParents.forEach((parent) => {
        parent.children.forEach((child, index) => {
            var _a, _b, _c;
            if (child.type !== 'citeGroup')
                return;
            const textChild = parent.children[index + 1];
            if ((textChild === null || textChild === void 0 ? void 0 : textChild.type) !== 'text')
                return;
            if (!((_a = textChild.value) === null || _a === void 0 ? void 0 : _a.match(/^\s*([,;]|[,;]{0,1}\s*(and))\s*$/)))
                return;
            const nextChild = parent.children[index + 2];
            if ((nextChild === null || nextChild === void 0 ? void 0 : nextChild.type) !== 'citeGroup')
                return;
            nextChild.children = [...((_b = child.children) !== null && _b !== void 0 ? _b : []), ...((_c = nextChild.children) !== null && _c !== void 0 ? _c : [])];
            nextChild.kind = child.kind;
            child.type = '__delete__';
            textChild.type = '__delete__';
        });
    });
    remove(tree, '__delete__');
}
/**
 * Remove parentheses surrounding citeGroup node and convert to parenthetical
 */
function removeCiteParentheses(tree) {
    const citeGroupParents = selectAll(':has(> text + citeGroup + text)', tree);
    citeGroupParents.forEach((parent) => {
        parent.children.forEach((child, index) => {
            var _a, _b, _c;
            if (child.type !== 'text' || !((_a = child.value) === null || _a === void 0 ? void 0 : _a.match(/[([]$/)))
                return;
            const closer = child.value[child.value.length - 1] === '(' ? ')' : ']';
            const citeChild = parent.children[index + 1];
            if ((citeChild === null || citeChild === void 0 ? void 0 : citeChild.type) !== 'citeGroup')
                return;
            const nextChild = parent.children[index + 2];
            if (nextChild.type !== 'text' || ((_b = nextChild.value) === null || _b === void 0 ? void 0 : _b[0]) !== closer)
                return;
            child.value = child.value.slice(0, -1);
            if (!child.value)
                child.type = '__delete__';
            citeChild.kind = 'parenthetical';
            (_c = citeChild.children) === null || _c === void 0 ? void 0 : _c.forEach((cite) => {
                if (cite.type === 'cite')
                    cite.kind = 'parenthetical';
            });
            nextChild.value = nextChild.value.slice(1);
            if (!nextChild.value)
                nextChild.type = '__delete__';
        });
    });
    remove(tree, '__delete__');
}
/**
 * When citations are separated by a hyphen, expand to fill in all intermediate citations
 */
function expandHyphenatedCites(tree, referenceList) {
    const citeGroupParents = selectAll(':has(> citeGroup + text + citeGroup)', tree);
    citeGroupParents.forEach((parent) => {
        parent.children.forEach((child, index) => {
            var _a, _b, _c, _d, _e;
            if (child.type !== 'citeGroup')
                return;
            if (((_a = child.children) === null || _a === void 0 ? void 0 : _a.length) !== 1)
                return;
            const firstCite = (_b = child.children) === null || _b === void 0 ? void 0 : _b[0];
            const textChild = parent.children[index + 1];
            if ((textChild === null || textChild === void 0 ? void 0 : textChild.type) !== 'text')
                return;
            if (!((_c = textChild.value) === null || _c === void 0 ? void 0 : _c.match(/^\s*[–-]\s*$/)))
                return;
            const nextChild = parent.children[index + 2];
            if ((nextChild === null || nextChild === void 0 ? void 0 : nextChild.type) !== 'citeGroup')
                return;
            if (((_d = nextChild.children) === null || _d === void 0 ? void 0 : _d.length) !== 1)
                return;
            const lastCite = (_e = nextChild.children) === null || _e === void 0 ? void 0 : _e[0];
            const firstInd = referenceList.indexOf(firstCite.label);
            const lastInd = referenceList.indexOf(lastCite.label);
            if (firstInd === -1 || lastInd === -1 || lastInd <= firstInd)
                return;
            const allCites = referenceList.slice(firstInd, lastInd + 1).map((citeLabel) => {
                var _a;
                const { label, identifier } = (_a = normalizeLabel(citeLabel)) !== null && _a !== void 0 ? _a : {};
                return { type: 'cite', kind: firstCite.kind, label: label !== null && label !== void 0 ? label : citeLabel, identifier };
            });
            nextChild.children = allCites;
            nextChild.kind = child.kind;
            child.type = '__delete__';
            textChild.type = '__delete__';
        });
    });
    remove(tree, '__delete__');
}
export function inlineCitationsTransform(tree, referenceIds) {
    let before = '';
    let current = JSON.stringify(tree);
    // Keep running these transforms until the tree no longer changes
    while (before !== current) {
        before = current;
        removeCiteChildren(tree);
        allCitesToCiteGroups(tree);
        flattenNestedCiteGroups(tree);
        combineAdjacentCiteGroups(tree);
        removeCiteSeparators(tree);
        expandHyphenatedCites(tree, referenceIds);
        removeCiteParentheses(tree);
        current = JSON.stringify(tree);
    }
}
